// SList.h - template created with new.header

#ifndef SList_h
#define SList_h

#include "slist.h"

class SList
{ 
  char** s;
  public:
    SList() {
      s = slist_new();
    }
    ~SList() {
      slist_delete(&s);
    }
    char** base() { return s; }
    char*** pbase() {return &s;}
    
    /// returns string at index
    char* get(int i) {return slist_get(&s, i);}

    /// adds a new items at the end of the list
    void append(const char* str) {slist_append(&s, str);}
    
    /// returns number of items in the list
    int count() {return slist_count(s);}
    
    /// removes items, does not resize or delete the base
    void clear() {slist_clear(&s);}

    /// inserts new item at index
    void insert(const char* str, int index) {slist_insert(&s, str, index);}

    /// adds other list to this, does not delete list that's added.
    /// plist2 can be any standard null terminated arg list or slist.
    void addList(SList* list) {slist_addList(&s, list->pbase());}
    void addList(char*** plist) {slist_addList(&s, plist);}

    /// removes items from index to end of list
    void trim(int index) {slist_trim(&s, index);}

    /// removes first n items from list
    void shift(int n) {slist_shift(&s, n);}

    /// removes item at i from list.
    void remove(int i) {slist_remove(&s, i);}

    /// reads lines (or args list) from a file, returns number of lines
    int fileRead(const char* filename) {return slist_fileRead(&s, filename);}
    
    /// writes lines (or args list) to a file, return 0 on success,
    /// list doesn't have to be a literal slist object, any null 
    /// terminated list not having newlines at end of lines will work.
    /// Optionally add a custom line_end string such as "\r\n" to each line.
    void fileWrite(const char* filename, const char* line_end = 0) 
      {slist_fileWrite(&s, filename, line_end);}

    /// Loads list with strings from a piped command,
    /// returns 0 on success
    int pipeRead(const char* pipecmd) {return slist_pipeRead(&s, pipecmd);}

    /// writes list through pipe, returns 0 on success.
    int pipeWrite(const char* syscmd) {return slist_pipeWrite(&s, syscmd);}

};

#endif // SList_h

