// Here are all the definitions for the def_loader

#define PARSER2_NO_WARN
#include "../lib/parser2.h"
#include "../lib/SList.h"
#include "../lib/lib.cpp"

#include "version.h"
#include "dict-util.h" // show_command_help
#include "ui-tool.h"    // usage()

#define obj (*g_parser_obj)

extern "C"
{

// finds the open and close of the tag in line and copies 
// what's between to a buffer, returns true on success.
bool parse_tag(char* buf, const char* tag, const char* line)
{
  bool ret = false;
//  PARSER_OBJ &prev;
//  parser_init(
//    &o,               /// addr of object to use
//    line,             /// initialized text to parse
//    strlen(line) + 1, /// exact length including terminating null
//              false      /// whether or not we can free inbuffer when done
//  );
parser_new(line, strlen(line), false);

  // .* <tag><VALUE</tag> .*
  while(1)
  {
    SAVE_PTRS();
    if(is_eoi())
    {
      obj.state = false;
      break;
    }
    do
    {
      (
        skip("<") 
        && skip(tag)
        && skip(">")
        && copy_to("</")
        && skip("</")
        // value_len = ip - value // but we don't need it
        && skip(tag)
        && skip(">")
        );
        
        if(!obj.state) break;
                    
                    // all tests passed
        strcpy(buf, obj.outbuf);
    }while(0); // unnest
    
    if(obj.state) break;
                    else
                    {
                      RESTORE_PTRS();
                      skip_nchars(1);
                    }
  }// while 1
  ret = obj.state;
//  parser_free(&o, prev);
  parser_delete(&obj);
  return ret;
}

// sets pointers to the lines with the tags <slots> and </slots>
// and returns true if successful, else false.
bool find_slots(SList* pxml, int* ptag_start, int* ptag_end)
{
  int start = -1, end = -1;
  
  for(int i = 1; i < pxml->count(); i++)
  {
    char* line = pxml->get(i);
    if(strstr(line, "<slots>"))
      start = i;
    else if(strstr(line, "</slots>"))
    {
      if(start < 0)
        break; // error, forget it.
        end = i;
      break;
    }
  }
  *ptag_start = start;
  *ptag_end = end;  
  
  if((start | end) < 0)
    return false;
  else
    return true;
}

// slots are all continguous inside the <slots> ... </slots> tags
static bool _get_slots(const char* filename, SList* pslots)
{
  bool ok;
  char buf[256];
  SList tmp;
  tmp.fileRead(filename);
  if(!tmp.count())
    return false;
  
  int tag_start, tag_end;
  ok = find_slots(&tmp, &tag_start, &tag_end);
  if(!ok)
    //    return false;
  return true; // no slots, not an error -rs
  
  for(int i = tag_start + 1; i < tag_end; i++)
  {
    char* line = tmp.get(i);
    if(parse_tag(buf, "slot", line))
      pslots->append(buf);
  }
  return true;
}


// signals can be anywhere in the file and may be duplicated several times.
static bool _get_signals(const char* filename, SList *s, SList *signal)
{
  char buf[256];
  s->fileRead(filename);
  if(!s->count())
    return false;
  
  for(int i = 1; i < s->count(); i++)
  {
    char* line = s->get(i);
    if(parse_tag(buf, "signal", line))
    {
      // find in dup in string
      bool state = true;
      for(int ii = 0; ii < signal->count(); ii++)
      {
        if(strcmp(signal->get(ii), buf) == 0)
        {
          state = false;
          break;
        }
      }
      if(state)
        signal->append(buf);
    }
  }
  return true;
}


bool moveSlots(SList* pslots, SList* pxml)
{
  bool ok;
  int tag_start, tag_end;
  ok = find_slots(pxml, &tag_start, &tag_end);
  if(!ok)
    // return false;
  return true; // no slots, not an error -rs
  
  int first_tag = tag_start + 1;
  for(int i = first_tag; i < tag_end; i++)
  {
    // continue moving and removing until count is done
    pslots->append(pxml->get(first_tag));
    pxml->remove(first_tag);
  }
  return true;
}

bool insertSlots(SList* pxml, SList* pslots)
{
  // pxml MUST still have the tags for the slots whether or not
  // the slots have been moved/removed.
  bool ok;
  int tag_start, tag_end;
  ok = find_slots(pxml, &tag_start, &tag_end);
  if(!ok)
    //return false;
  return true; // no slots, not an error -rs
  
  int insloc = tag_start + 1; // where we'll be inserting
  for(int i = pslots->count() -1; i >= 0; i--)
    pxml->insert(pslots->get(i), insloc);
  return true;
}

bool sortSlots(const char* filename)
{
// #warning user is responsible for backup of original file in case it
// gets corrupted.
  bool ret;
  char swapbuf[256];
  SList xnames;
  SList xml;
  SList names;
  
  xml.fileRead(filename);
  if(!xml.count())
    return false;
  
  ret = moveSlots(&xnames, &xml);
  if(!ret)
    return false;
  
  // xnames is the xml containing the names
    
    ret = _get_slots(filename, &names);
    // names is the alpha-sortable string
    
    // bool find_slots(SList* pxml, int* ptag_start, int* ptag_end)
    bool changed = true;
    while(changed)
    {
      changed = false;
      for(int i = 1; i < names.count(); i++)
      {
        if(strcmp(names.get(i), names.get(i-1)) < 0) // swap 'em?
      {
        // swap both sets of names
        names.insert(names.get(i), i-1);
        names.remove(i+1);
        xnames.insert(xnames.get(i), i-1);
        xnames.remove(i+1);
        changed = true;
      }
      }
    }
    
    ret = insertSlots(&xml, &xnames);
    if(!ret)
      return false;
    
    //  xml.fileWrite(filename);
      for(int i = 0; i < xml.count(); i++) 
        printf("%s\n", xml.get(i));
      
      return true;
}

// input is a slot name as output from listSlots, with no
// tags and no parameter names.
static bool _xlate_slot(char* buf, int bufmax, const char* slotname)
{
  bool ret;
  char param_name[256];
  int param_id = 'A';
        parser_new(slotname, strlen(slotname), false);
        do
        {
          copy_to("(");
          if(!obj.state) break;
                      copy("(");
          if(is(")")) // empty list
    {
      copy(")");
      break; // true
    }
    // else has parameters
    // for every , and ')' insert a space and a param name
    while(1)
    {
      if(is_eoi())
      {
        obj.state = false;
        break;
      }
      if(copy_to(","))
      {
        skip(",");
        obj.op += sprintf(obj.op, " arg%c,", param_id++);
        // and continue
      }
      
      else if(copy_to(")"))
      {
        skip(")");
        // let caller decide whether to use the trailing semicolon -rs
        // obj.op += sprintf(obj.op, " arg%c);", param_id++);
        obj.op += sprintf(obj.op, " arg%c)", param_id++);
        
        // break
        break;
      }
      else
        obj.state = false;
      break;
    }
        }while(0);
        if((ret = obj.state))
        {    
          strcpy(buf, obj.outbuf);
        }
        ret = obj.state;
        parser_delete(&obj);
        return ret;
}


static bool _get_classname(const char* filename, SList* c)
{
  char buf[256];
  SList x;
  x.fileRead(filename);
  
  // quick and dirty, really needs more checking TODO:
  char* line = x.get(2); // <class>Name</class>
  char* op;
  char* ip = strstr(line, "<class>");
  if(!ip)
    return false;
  
  ip += 7;
  op = buf;
  while(*ip != '<')
    *op++ = *ip++;
  *op = 0;
  c->append(buf);
  return true;
}

bool writeCpp(const char* filename)
{
  bool ok;
  SList s;
  SList c;
  
  ok = _get_slots(filename, &s);
  if(!ok)
    return false;
  
  ok = _get_classname(filename, &c);
  if(!ok)
    return false;
  
  /////////////////////////////////////
    // c@0 = classname
  char* p;
  char bname[256];
  sprintf(bname, filename);
  p = strrchr(bname, '.');
  if(p) *p = 0;
  printf("\n// Place in %s.cpp\n", bname);
  for(int i = 0; i < s.count(); i++)
  {
    char slotname[256];
    _xlate_slot(slotname, 256, s.get(i));
    printf("void %s::%s\n", c.get(0), slotname);
    printf("{\n#warning Not Implemented Yet!\n}\n\n");
  }  
  return true;
}


bool writeH(const char* filename)
{
  //  #warning writes slots per uiname.h
  bool ok;
  SList s;
  ok = _get_slots(filename, &s);
  if(!ok)
    return false;
  
  /////////////////////////////////////
    char* p;
    char bname[256];
    sprintf(bname, filename);
    p = strrchr(bname, '.');
    if(p) *p = 0;
                      printf("\n// Place below Q_OBJECT in %s.h\n", bname);
    printf("\npublic slots:\n");  
        for(int i = 0; i < s.count(); i++)
        {
          char slotname[256];
          _xlate_slot(slotname, 256, s.get(i));
          printf("  void %s;\n", slotname);
        }  
        return true;
}


// just show the list of slots in the ui file.
bool listSlots(const char* filename)
{
  bool ok;  
  SList s;
  ok = _get_slots(filename, &s);
  if(ok)
  {
    for(int i = 0; i < s.count(); i++)
      printf("%s\n", s.get(i));
  }
  return ok;
}


bool listSignals(const char* filename)
{
  bool ret;
  SList s;
  SList signal;
  ret = _get_signals(filename, &s, &signal);
  if(ret)
  {
    for(int i = 0; i < signal.count(); i++)
      printf("%s\n", signal.get(i));
  }
  return ret;
}

//DEf(uicables, -1)
bool uicAble(const char* opt_dirname)
{
  char cmd[256];
  const char* dirname;
  SList files;
//  SList s;
  
  if(opt_dirname == 0)
    dirname = ".";
  else
    dirname = opt_dirname;
  
  // ls dirname/*.ui 2>/dev/null
  sprintf(cmd, "ls %s/*.ui 2>/dev/null", dirname);
  files.pipeRead(cmd);
  if(files.count() == 0)
    return true; // not an error
  
  for(int i = 0; i < files.count(); i++)
  {
    char* filename = strdup(files.get(i));
    char* basename = strrchr(filename, '/');
    if(basename)
      basename++;
    else
      basename = filename;
    char* p = strchr(basename, '.');
    if(p)
      *p = 0;
    // the -o switch works right with a ./filename, but > doesn't.  bug: uic writes
    // the #ifdef header name without the UI_ prefix.  So we do this:
    printf("uic %s/%s.ui -o %s/ui_%s.h\n", dirname, basename, dirname, basename);
  }
  return true;
}


//DEF(mocable, -1)
bool mocAble(const char* opt_dirname)
{
  char cmd[256];
  const char* dirname;
  SList files;
//  SList s;
  
  if(opt_dirname == 0)
    dirname = ".";
  else
    dirname = opt_dirname;
  
  // ls dirname/*.ui 2>/dev/null
  sprintf(cmd, "ls %s/*.h 2>/dev/null", dirname);
  files.pipeRead(cmd);
  if(files.count() == 0)
    return true; // not an error
  
  for(int i = 0; i < files.count(); i++)
  {    
    char* filename = strdup(files.get(i));
    
    // see if it has a Q_OBJECT
    // grep Q_OBJECT filename
    sprintf(cmd, "grep Q_OBJECT %s >/dev/null 2>&1", filename);
    if(system(cmd))
      continue;

    char* basename = strrchr(filename, '/');
    if(basename)
      basename++;
    else
      basename = filename;
    char* p = strchr(basename, '.');
    if(p)
      *p = 0;
    
    sprintf(cmd, "moc %s/%s.h > /dev/null 2>&1", dirname, basename);
    if(system(cmd) == 0)
    {
      strchr(cmd, '>')[0] = 0;
      printf("%s -o %s/moc_%s.cpp\n", cmd, dirname, basename);
    }
      
  }
  return true;
}

//DEF(rccable, -1)
bool rccAble(const char* opt_dirname)
{
  char cmd[256];
  const char* dirname;
  SList files;
//  SList s;
  
  if(opt_dirname == 0)
    dirname = ".";
  else
    dirname = opt_dirname;
  
  // ls dirname/*.ui 2>/dev/null
  sprintf(cmd, "ls %s/*.qrc 2>/dev/null", dirname);
  files.pipeRead(cmd);
  if(files.count() == 0)
    return true; // not an error
  
  for(int i = 0; i < files.count(); i++)
  {
    char* filename = strdup(files.get(i));
    char* basename = strrchr(filename, '/');
    if(basename)
      basename++;
    else
      basename = filename;
    char* p = strchr(basename, '.');
    if(p)
      *p = 0;
    
    sprintf(cmd, "rcc %s/%s.qrc > /dev/null 2>&1", dirname, basename);
    if(system(cmd) == 0)
    {
      strchr(cmd, '>')[0] = 0;
      printf("%s -o %s/qrc_%s.cpp\n", cmd, dirname, basename);
    }
  }
  return true;
}


//DEF(qtCommands, -1)
bool qtCommands(const char* opt_dirname)
{
  uicAble(opt_dirname);
  mocAble(opt_dirname);
  rccAble(opt_dirname);  
  return true;
}

extern int g_argc;
extern char** g_argv;
extern int ui2cpp_main(int argc, char** argv); // main wrapper

// wrapper for imported ui2cpp main()
bool ui2cpp()
{
  return 0 == ui2cpp_main(g_argc -1, g_argv +1);
}

bool printVersion()
{
  printf("ui-tool version %s\n", version);
  return true;
}

bool commands(const char* opt_substr)
{
  show_command_help(opt_substr);
  return true;
}

bool help()
{
  usage(0);
  return true;
}

} // extern C
